(* Implémentation de files de priorité par des tas *)

(* Les priorités sont des entiers, 
	 la priorité la plus petite est la plus pressante 
*)
type 'a elem = {element : 'a; priorite : int}

(* Le type des files de priorité par tas.
	 
	 Le champ tas est le tas lui-même (ordonné suivant les priorité des 'a elem).
	 Le champ longueur est le nombre d'éléments du tas. 
*)
type 'a tas_min = {
	tas : 'a elem array;
	mutable longueur : int (* nb d'objets stockés dans l'arbre *)
}
type 'a t = 'a tas_min

exception Empty
exception Full

(* Type qui sert à rendre le code plus lisible *)
type indice = int

(* Trois fonctions auxiliaires fort fort fort pratiques *)
let parent (i : indice) : indice = (i-1)/2
let enf_gauche (i : indice) : indice = 2*i+1
let enf_droite (i : indice) : indice = 2*i+2


(* Crée un tas vide. 
	 x est nécessaire pour initialiser l'Array sous-jacent. 
*)
let make (longueur_max : int) (x : 'a) : 'a tas_min =
	if longueur_max < 0 || longueur_max > Sys.max_array_length then
		raise (Invalid_argument "Prio.make : invalid length")
	else
		{tas = Array.make longueur_max {element = x; priorite = 666}; longueur = 0}


(* Teste si un tas est vide *)
let is_empty (t : 'a tas_min) : bool =
	t.longueur = 0

(* Renvoie lenombre d'éléments d'une file *)
let length (t : 'a tas_min) : int =
	t.longueur


(* Réalise une percolation vers le haut 
	 de l'élément d'indice start dans le tas t.

	 Lève Invalid_argument si start n'est pas un indice valide.
*)
let percole_haut (t : 'a tas_min) (start : indice) : unit =
	assert( 0 <= start && start < t.longueur );

	let tas = t.tas in
	let i = ref start in

	(* tant que le parent est plus grand, percoler avec lui *)
	while !i <> 0 && tas.(!i).priorite < tas.(parent !i).priorite do
		let temp = tas.(!i) in
		tas.(!i) <- tas.(parent !i);
		tas.(parent !i) <- temp;
		i := parent !i
	done


(* Réalise une percolation vers le bas 
	 de l'élément d'indice start dans le tas t.

	 Lève Invalid_argument si start n'est pas un indice valide.
	 Cette exception ne devrait pas 
*)
let percole_bas (t : 'a tas_min) (start : indice) : unit = 
	assert( 0 <= start && start < t.longueur );

	let tas = t.tas in
	let i = ref start in
	let fini = ref false in
	
	(* tant qu'un enfant est plus petit, percoler avec le min des enfants.
				On doit distinguer 3 cas : pas d'enfant, 
																	uniquement enfant gauche, 
																	deux enfants
	*)
	while not !fini do
		let eg = enf_gauche !i in
		let ed = enf_droite !i in

		(* Pas d'enfant -> percolation finie *)
		if eg >= t.longueur then
			fini := true

		(* un seul enfant -> on essaye de percoler avec lui.
				Si on ne peut pas,la percolation est finie. *)
		else if ed >= t.longueur then begin
			if tas.(!i).priorite > tas.(eg).priorite then begin
				let temp = tas.(!i) in
				tas.(!i) <- tas.(eg);
				tas.(eg) <- temp;
				i := eg
			end
			else
				fini := true
		end

		(* deux seul enfant -> on essaye de percoler avec le min des enfants.
				Si on ne peut pas,la percolation est finie. *)
		else begin
			let e_min = if tas.(eg).priorite < tas.(ed).priorite then eg else ed in
			if tas.(!i).priorite > tas.(e_min).priorite then begin 
				let temp = t.tas.(!i) in
				t.tas.(!i) <- t.tas.(e_min);
				t.tas.(e_min) <- temp;
				i := e_min
			end
			else 
				fini := true
		end

	done


(* Insère l'élément x dans le tas t avec la priorité p.
	 Lève Full si le tas est plein.
*)
let add (t : 'a tas_min) (x : 'a) (p : int) : unit =
	if t.longueur = Array.length t.tas then 
		raise Full
	else begin
		t.longueur <- t.longueur + 1;
		t.tas.(t.longueur - 1) <- {element = x; priorite = p};
		percole_haut t (t.longueur - 1)
	end


(* Renvoie l'élément minimal du tas t et sa priorité. 
	 Ne modifie pas le tas.
	 Lève Empty si le tas est vide.		
*)
let peek (t : 'a tas_min) : 'a * int =
	if t.longueur = 0 then 
		raise Empty
	else
		(t.tas.(0).element, t.tas.(0).priorite)


(* Comme peek mais avec une option et non une exception *)
let peek_opt (t : 'a tas_min) : ('a * int) option =
	if t.longueur = 0 then 
		None
	else
		Some (t.tas.(0).element, t.tas.(0).priorite)


(* Extrait et renvoie l'élément minimal du tas t.
	 Modifie le tas.
	 Lève Empty si le tas est vide.		
*)
let take (t : 'a tas_min) : 'a * int =
	if t.longueur = 0 then 
		raise Empty
	else begin
		let mini = t.tas.(0) in
		t.longueur <- t.longueur-1;

		(* Remonter la dernière feuille et la percoler vers le bas,
			 sauf si le tas est maintenant vide *)
		if t.longueur > 0 then begin	
			t.tas.(0) <- t.tas.(t.longueur);
			percole_bas t 0
		end;

		(mini.element, mini.priorite)
	end


(* Comme take mais avec une option et non une exception. *)
let take_opt (t : 'a tas_min) : ('a * int) option =
	if t.longueur = 0 then 
		None
	else begin
		let mini = t.tas.(0) in
		t.longueur <- t.longueur-1;

		(* Remonter la dernière feuille et la percoler vers le bas,
			 sauf si le tas est maintenant vide *)
		if t.longueur > 0 then begin	
			t.tas.(0) <- t.tas.(t.longueur);
			percole_bas t 0
		end;

		Some (mini.element, mini.priorite)
	end