(* Implémentation de files de priorité avec des tas de longueur bornée. *)

(** Type des files de priorité contenant des éléments de type 'a.

    Les priorités sont des entiers. Un entier plus petit correspond
    à une priorité plus importante.
*)
type 'a t


(** Levé quand Prio.add est appliqué à une file 
    ayant atteint sa capacité maximale. *)
exception Empty

(** Levé quand Prio.peek ou Prio.take est appliqué à une file vide. *)
exception Full


(** make len x   crée une file de priorité vide qui pourra contenir 
    au maximum len éléments. L'élément x fourni n'a pas d'importance.

    Lève Invalid_argument si len < 0. 
*)
val make : int -> 'a -> 'a t


(** is_empty f   renvoie true si et seulement si f ne contient aucun élément *)
val is_empty : 'a t -> bool

(** length f   renvoie le nombre d'éléments dans la file de priorité f *)
val length : 'a t -> int


(** add f x p    insère l'élément x dans la file f avec la priorité p.
    f est donc modifiée.

    Lève Prio.Full si f avait déjà atteint sa capacité maximale.
*)
val add : 'a t -> 'a -> int -> unit

(** peek f    renvoie la paire (élément, priorité) de priorité
    minimale dans f (c'est la prochaine à sortir de la file).

    Ne modifie pas f.

    Lève Prio.Empty si la file est vide.
*)
val peek : 'a t -> 'a * int

(** peek_opt f    renvoie  Some (élément, priorité) où cette paire est
    de priorité minimale dans f (c'est la prochaien à sortir de la file).
    Renvoie None si la file est vide.

    Ne modifie pas f.

    Équivalent à   try Some (peek f) with Prio.Empty -> None
*)
val peek_opt : 'a t -> ('a * int) option

(** take f x p    extrait et renvoie la paire (élément, priorité) de priorité
    minimale dans f (c'est à dire la prochaine à sortie de la file).
    f est donc modifiée.

    Lève Prio.Empty si la file est vide.
*)
val take : 'a t -> 'a * int

(** take_opt f x p    extrait et renvoie Some (élément, priorité) où cette paire
    est de priorité minimale dans f (c'est à dire la prochaine à 
    sortir de la file). Renvoie None si la file est vide.

    f est donc modifiée.

    Équivalent à   try Some (take f) with Prio.Empty -> None
*)
val take_opt : 'a t -> ('a * int) option
